#!/usr/bin/tclsh
# Part of MCU 8051 IDE ( http://mcu8051ide.sf.net )

############################################################################
#    Copyright (C) 2007-2009 by Martin Ošmera                              #
#    martin.osmera@gmail.com                                               #
#                                                                          #
#    This program is free software; you can redistribute it and#or modify  #
#    it under the terms of the GNU General Public License as published by  #
#    the Free Software Foundation; either version 2 of the License, or     #
#    (at your option) any later version.                                   #
#                                                                          #
#    This program is distributed in the hope that it will be useful,       #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#    GNU General Public License for more details.                          #
#                                                                          #
#    You should have received a copy of the GNU General Public License     #
#    along with this program; if not, write to the                         #
#    Free Software Foundation, Inc.,                                       #
#    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             #
############################################################################

# --------------------------------------------------------------------------
# DESCRIPTION
# Background error handler
# --------------------------------------------------------------------------

namespace eval ErrorHandler {
	variable num_of_opended	0	;# Int: Number of currently opened dialogs
	variable count		0	;# Int: Counter of ivokations
	variable enabled	1	;# Bool: Dialog window enabled

	## Open dialog window
	 # @parm String message - Error message
	 # @return void
	proc open_dialog {message} {
		variable count		;# Int: Counter of ivokations
		variable enabled	;# Bool: Dialog window enabled
		variable num_of_opended	;# Int: Number of currently opened dialogs

		if {$num_of_opended > 2} {
			puts stderr "ERROR MESSAGE SUPPRESED (too many error dialogs opened at the time)"
			return
		}
		incr num_of_opended

		# Send error message to standard error output
		puts stderr [string repeat # 64]
		puts stderr "#                       PROGRAM ERROR                          #"
		puts stderr [string repeat # 64]
		puts stderr $::errorInfo
		puts stderr [string repeat # 64]

		# Save log file
		if {![catch {set log_file [open ~/mcu8051ide_errors.log a]}]} {
			puts $log_file [string repeat # 64]
			puts $log_file "Program version:\t${::VERSION}"
			puts $log_file "Tcl version:\t\t${::tcl_version}"
			puts $log_file "Tk version:\t\t${::tk_version}"
			puts $log_file [string repeat - 64]
			puts $log_file $::errorInfo
			close $log_file
		}

		# Create dialog window
		if {!$enabled} {return}
		incr count
		set win [toplevel .error_dialog$count -bg {#EE0000} -class {Error message}]

		# Create window frames
		set main_frame [frame $win.main_frame]
		set top_frame [frame $main_frame.top_frame -bg {#EE0000}]
		set middle_frame [frame $main_frame.middle_frame]
		set bottom_frame [frame $main_frame.bottom_frame]

		# Create window header
		pack [label $top_frame.header_lbl	\
			-text "PROGRAM ERROR  "		\
			-bg {#EE0000} -fg {#FFFFFF}	\
			-font [font create		\
				-family helvetica	\
				-size -24		\
				-weight bold		\
			]	\
		] -side left -fill x -expand 1

		# Create error message text and scrollbar
		pack [text $middle_frame.text				\
			-bg {white} -bd 0				\
			-yscrollcommand "$middle_frame.scrollbar set"	\
			-width 0 -height 0 -relief flat -wrap word	\
		] -side left -fill both -expand 1 -padx 5 -pady 5
		bind $middle_frame.text <Button-1> {focus %W}
		pack [scrollbar $middle_frame.scrollbar		\
			-orient vertical			\
			-command "$middle_frame.text yview"	\
		] -fill y -side right

		# Create text tags
		$middle_frame.text tag configure tag_bold	\
			-font [font create			\
				-family {courier}		\
				-weight bold			\
				-size -14			\
			]
		$middle_frame.text tag configure tag_big_bold	\
			-font [font create			\
				-family {helvetica}		\
				-weight bold			\
				-size -17			\
			]

		# Write error message
		$middle_frame.text insert end "Error log saved in: $::env(HOME)/mcu8051ide_errors.log.\n"
		$middle_frame.text insert end "Please send this file to <martin.osmera@gmail.com> or report this bug at http://mcu8051ide.sf.net/forum\n\n\n"
		$middle_frame.text tag add tag_big_bold 1.0 3.0
		$middle_frame.text insert end "ERROR DETAILS:\n--------------\n"
		$middle_frame.text tag add tag_bold 5.0 7.0
		$middle_frame.text insert end $::errorInfo
		$middle_frame.text configure -state disabled

		# Create button frame
		pack [button $bottom_frame.skip				\
			-text "Skip errors" -compound left -bd 1	\
			-command "
				set ::ErrorHandler::enabled 0
				::ErrorHandler::close_dialog $count
			"	\
		] -side left
		pack [button $bottom_frame.ok			\
			-text "Close" -bg {#88FF88} -padx 50 	\
			-activebackground {#AAFFAA} -bd 1	\
			-command "::ErrorHandler::close_dialog $count"	\
		] -side right
		focus -force $bottom_frame.ok
		
		# Pack window frames
		pack $top_frame -fill x -anchor n
		pack $middle_frame -fill both -expand 1
		pack $bottom_frame -fill x
		pack $main_frame -fill both -expand 1 -padx 5 -pady 5

		# Configure dialog window
		set x [expr {[winfo screenwidth $win] / 2 - 225}]
		set y [expr {[winfo screenheight $win] / 2 - 125}]
		wm iconphoto $win ::ICONS::16::bug
		wm title $win "PROGRAM ERROR - MCU 8051 IDE"
		wm minsize $win 450 250
		wm geometry $win =550x250+$x+$y
		wm protocol $win WM_DELETE_WINDOW "::ErrorHandler::close_dialog $count"
		update
		raise $win
		catch {grab $win}
	}

	## Close dialog window
	 # @parm Int number - Dialog unique number
	 # @return void
	proc close_dialog {number} {
		variable num_of_opended	;# Int: Number of currently opened dialogs

		incr num_of_opended -1
		destroy .error_dialog$number
	}
}

# Register error handler
proc bgerror {message} {
	::ErrorHandler::open_dialog $message
}
