#!/usr/bin/tclsh
# Part of MCU 8051 IDE ( http://mcu8051ide.sf.net )

############################################################################
#    Copyright (C) 2007-2009 by Martin Ošmera                              #
#    martin.osmera@gmail.com                                               #
#                                                                          #
#    This program is free software; you can redistribute it and#or modify  #
#    it under the terms of the GNU General Public License as published by  #
#    the Free Software Foundation; either version 2 of the License, or     #
#    (at your option) any later version.                                   #
#                                                                          #
#    This program is distributed in the hope that it will be useful,       #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#    GNU General Public License for more details.                          #
#                                                                          #
#    You should have received a copy of the GNU General Public License     #
#    along with this program; if not, write to the                         #
#    Free Software Foundation, Inc.,                                       #
#    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             #
############################################################################


# --------------------------------------------------------------------------
# DESCRIPTION
#
# --------------------------------------------------------------------------

class LedDisplay {
	inherit VirtualHWComponent

	common cb_font	[font create -weight bold -size -10 -family {helvetica}]
	common COMPONENT_NAME	"LED Display"
	common CLASS_NAME	"LedDisplay"
	common COMPONENT_ICON	{leddisplay}

	# red orange yellow   green blue purple
	common COLORS {
		{#AA5555 #FF0000}
		{#AAAA55 #FF8800}
		{#AAAA55 #FFFF00}

		{#55AA55 #00FF00}
		{#5555AA #0000FF}
		{#AA55AA #8800FF}
	}

	# Configuration menu
	common CONFMENU {
		{cascade	{Color}			0	"colorize"	.color	false 1 {
			{radiobutton	"Red"		{}
				::LedDisplay::color		{red}
				"color_changed"	0	""}
			{radiobutton	"Orange"	{}
				::LedDisplay::color		{orange}
				"color_changed"	0	""}
			{radiobutton	"Yellow"	{}
				::LedDisplay::color		{yellow}
				"color_changed"	0	""}
			{radiobutton	"Green"		{}
				::LedDisplay::color		{green}
				"color_changed"	0	""}
			{radiobutton	"Blue"		{}
				::LedDisplay::color		{blue}
				"color_changed"	0	""}
			{radiobutton	"Purple"	{}
				::LedDisplay::color		{purple}
				"color_changed"	0	""}
		}}
		{separator}
		{command	{Show help}		{}	5	"show_help"	{help}
			"Show brief help"}
		{separator}
		{command	{Save configuration}	{}	0	"save_as"	{filesave}
			"Save configuration into a file"}
		{command	{Load configuration}	{}	0	"load_from"	{fileopen}
			"Load configuration from a file"}
	}
	
	private variable conf_led_color		{red}

	private variable leds
	private variable wires
	private variable connection_port
	private variable connection_pin
	
	constructor {_project} {
		set component_name	$COMPONENT_NAME
		set class_name		$CLASS_NAME
		set component_icon	$COMPONENT_ICON

		set project $_project
		
		$project pale_register_output_device $this
		$project pale_set_modified

		array set connection_port	{0 - 1 - 2 - 3 - 4 - 5 - 6 - 7 -}
		array set connection_pin	{0 - 1 - 2 - 3 - 4 - 5 - 6 - 7 -}

		create_gui
		mcu_changed
		on_off [$project pale_is_enabled]

		for {set i 0} {$i < 8} {incr i} {
			$canvas_widget.cb_b$i setvalue first
			$canvas_widget.cb_p$i setvalue first
		}
	}

	destructor {
		$project pale_unregister_output_device $this
		destroy $win
	}

	private method create_gui {} {
		set win [toplevel .leddisplay$count -class $component_name]

		set canvas_widget [canvas $win.canvas	\
			-bg white -width 0 -height 0	\
			-highlightthickness 0		\
		]

		draw_8_segment 85 30
		draw_wires 0 -10

		set cb_p_x0 200
		set cb_b_x0 200
		set cb_p_x1 40
		set cb_b_x1 40
		set y_0 30
		set y_1 120
		set y_inc 30
		for {set i 0} {$i < 8} {incr i} {
			if {$i == 0} {
				set y $y_0
				set cb_p_x $cb_p_x0
				set cb_b_x $cb_b_x0
			} elseif {$i == 4} {
				set y $y_1
				set y_inc -$y_inc
				set cb_p_x $cb_p_x1
				set cb_b_x $cb_b_x1
			}

			$canvas_widget create window $cb_p_x $y -anchor e	\
				-window [ComboBox $canvas_widget.cb_p$i	\
					-width 1 -font $cb_font		\
					-modifycmd "$this reconnect $i"	\
					-entrybg white -editable 0	\
					-hottrack 1			\
				]
			$canvas_widget create window $cb_b_x $y -anchor w	\
				-window [ComboBox $canvas_widget.cb_b$i	\
					-width 1 -font $cb_font		\
					-modifycmd "$this reconnect $i"	\
					-values {- 0 1 2 3 4 5 6 7}	\
					-entrybg white -editable 0	\
					-hottrack 1			\
				]

			catch {
				bindtags $canvas_widget.cb_p$i.e	\
					[list $canvas_widget.cb_p$i.e Entry all .]
				bindtags $canvas_widget.cb_b$i.e	\
					[list $canvas_widget.cb_b$i.e Entry all .]
			}

			incr y $y_inc
		}
		
		set start_stop_button [Button $canvas_widget.start_stop_button	\
			-bd 1 -command "$this on_off_button_press"		\
			-padx 0 -pady 0 -highlightthickness 0 -width 3		\
			-helptext "Turn HW simulation on/off"			\
		]
		setStatusTip -widget $start_stop_button -text "Turn HW simulation on/off"
		bind $start_stop_button <Button-3> "$this on_off_button_press; break"
		$canvas_widget create window 115 2 -window $start_stop_button -anchor nw
		
		set conf_button [button $canvas_widget.conf_but	\
			-image ::ICONS::16::configure -bd 0	\
			-command "$this config_menu"		\
			-relief flat -overrelief raised		\
			-bg white -activebackground white	\
			-highlightthickness 0			\
		]
		setStatusTip -widget $conf_button -text "Configure"
		$canvas_widget create window 113 2 -window $conf_button -anchor ne


		$canvas_widget create text 40 5	\
			-text {PORT}			\
			-font $cb_font			\
			-anchor ne
		$canvas_widget create text 50 5	\
			-text {BIT}			\
			-font $cb_font			\
			-anchor nw
		$canvas_widget create text 200 5	\
			-text {PORT}			\
			-font $cb_font			\
			-anchor ne
		$canvas_widget create text 210 5	\
			-text {BIT}			\
			-font $cb_font			\
			-anchor nw

		$canvas_widget create text 5 155	\
			-text {Note}			\
			-font $cb_font			\
			-anchor w
		$canvas_widget create window 40 155		\
			-window [entry $canvas_widget.usr_note	\
				-bg white -validate all		\
				-vcmd "$this set_modified"	\
			]	\
			-width 180 -anchor w
		bindtags $canvas_widget.usr_note	\
			[list $canvas_widget.usr_note Entry $win all .]

		pack $canvas_widget -fill both -expand 1
		
		# Set window parameters
		wm geometry $win =230x170
		wm iconphoto $win ::ICONS::16::$component_icon
		wm title $win "$component_name - [string trim $project {:}] - MCU 8051 IDE"
		wm resizable $win 0 0
		wm protocol $win WM_DELETE_WINDOW "$this close_window"
		
		bindtags $win [list $win Toplevel all .]
	}

	public method reconnect {i} {
		set connection_port($i) [$canvas_widget.cb_p$i get]
		set connection_pin($i)	[$canvas_widget.cb_b$i get]
		if {$connection_pin($i) != {-}} {
			set connection_pin($i)	[expr {7 - $connection_pin($i)}]
		}

		if {$drawing_on} {
			new_state [$project pale_get_true_state]
		}
		set_modified
	}

	private method draw_wires {x y} {
		set coords {
			{
				120 40	180 40
			} {
				145 70	180 70
			} {
				140 100	180 100
			} {
				110 125	110 130	180 130
			} {
				87 110	82 110	82 130	50 130
			} {
				91 75	82 75	82 100	50 100
			} {
				116 78	116 70	50 70
			} {
				138 125	138 150	5 150	5 40	30 40
			}
		}

		for {set i 0} {$i < 8} {incr i} {
			set coordinates [list]
			set local_coords [lindex $coords $i]
			set len [llength $local_coords]

			for {set m 0; set n 1} {$n < $len} {incr m 2; incr n 2} {
				lappend coordinates	\
					[expr {[lindex $local_coords $m] + $x}]
				lappend coordinates	\
					[expr {[lindex $local_coords $n] + $y}]
			}

			set wires($i) [$canvas_widget create line	\
				$coordinates -width 1 -fill #000000	\
			]
		}
	}

	private method draw_8_segment {x y} {
		set coords {
			{
				19 7	25 1	47 1	53 7	53 8
				47 14	25 14	19 8
			} {
				55 9	62 16	58 34	50 42	44 36
				49 15	55 9
			} {
				50 45	57 52	53 70	46 77	45 77
				39 71	44 51	50 45
			} {
				15 73	38 73	44 79	37 86	15 86
				9 80	9 79
			} {
				7 78	15 70	19 52	12 45	5 52
				1 72
			} {
				12 42	20 34	25 16	17 9	10 16
				6 36
			} {
				14 43	20 37	42 37	48 43	48 44
				42 50	20 50	14 44
			}
		}

		for {set i 0} {$i < 7} {incr i} {
			set coordinates [list]
			set local_coords [lindex $coords $i]
			set len [llength $local_coords]

			for {set m 0; set n 1} {$n < $len} {incr m 2; incr n 2} {
				lappend coordinates	\
					[expr {[lindex $local_coords $m] + $x}]
				lappend coordinates	\
					[expr {[lindex $local_coords $n] + $y}]
			}

			set leds($i) [$canvas_widget create polygon	\
				$coordinates -width 0 -fill #888888	\
			]
		}
		set leds(7) [$canvas_widget create oval		\
			[expr {49 + $x}] [expr {77 + $y}]	\
			[expr {58 + $x}] [expr {86 + $y}]	\
			-width 0 -fill #888888			\
		]
		
		foreach coords {{35 7} {53 25} {48 62} {26 79} {10 61} {15 25} {31 43}} \
			text {A B C D E F G} \
		{
			$canvas_widget create text			\
				[expr {[lindex $coords 0] + $x}]	\
				[expr {[lindex $coords 1] + $y}]	\
				-text $text -fill {#FFFFFF}		\
				-font $::smallfont
		}
	}

	private method which_port_pin {i} {
		return [list $connection_port($i) $connection_pin($i)]
	}

	public method on_off_button_press {} {
		$project pale_all_on_off
	}


	# ------------------------------------------------------------------
	# VIRTUAL HW COMMON INTERFACE
	# ------------------------------------------------------------------

	public method mcu_changed {} {
		set avaliable_ports [concat - [$project pale_get_avaliable_ports]]
		
		for {set i 0} {$i < 8} {incr i} {
			$canvas_widget.cb_p$i configure -values $avaliable_ports

			if {[lsearch -ascii -exact $avaliable_ports $connection_port($i)] == -1} {
				$canvas_widget.cb_p$i setvalue first
				set connection_port($i) {-}
			}
		}
	}
	
	public method on_off {state} {
		set drawing_on $state
		
		if {$state} {
			$start_stop_button configure	-text {ON}			\
				-bg {#AAFFAA}		-activebackground {#DDFFDD}	\
				-fg {#000000}		-activeforeground {#00FF00}

			new_state [$project pale_get_true_state]

		} else {
			$start_stop_button configure	-text {OFF}			\
				-bg {#FFAAAA}		-activebackground {#FFDDDD}	\
				-fg {#000000}		-activeforeground {#FF0000}

			for {set i 0} {$i < 8} {incr i} {
				$canvas_widget itemconfigure $leds($i) -fill {#888888}
				$canvas_widget itemconfigure $wires($i) -fill {#000000}
			}
		}
	}

	public method new_state {state} {	
		set color_idx [lsearch -ascii -exact		\
			{red orange yellow green blue purple}	\
			$conf_led_color				\
		]

		for {set i 0} {$i < 8} {incr i} {
			set pp [which_port_pin $i]
			if {[lindex $pp 0] == {-} || [lindex $pp 1] == {-}} {
				$canvas_widget itemconfigure $leds($i) -fill {#888888}
				$canvas_widget itemconfigure $wires($i) -fill {#000000}
				continue
			}

			switch -- [lindex $state $pp] {
				{0} {
					set segment_color	{2}
					set wire_color		{#00FF00}
				}
				{1} {
					set segment_color	{0}
					set wire_color		{#FF0000}
				}
				{=} {
					set segment_color	{0}
					set wire_color		{#FF00AA}
				}
				{} {
					set segment_color	{0}
					set wire_color		{#000000}
				}
				{?} {
					set segment_color	{0}
					set wire_color		{#888888}
				}
				default {
					set segment_color	{1}
					set wire_color		{#FF8800}
				}
			}

			if {!$segment_color} {
				set segment_color {#888888}
			} {
				incr segment_color -1
				set segment_color [lindex $COLORS [list $color_idx $segment_color]]
			}

			$canvas_widget itemconfigure $leds($i) -fill $segment_color
			$canvas_widget itemconfigure $wires($i) -fill $wire_color
		}
	}

	public method withdraw_window {} {
		wm withdraw $win
	}

	public method get_config {} {
		return [list		\
			$class_name	\
			[list		\
				[array get connection_port]	\
				[array get connection_pin]	\
				[wm geometry $win]		\
				[$canvas_widget.usr_note get]	\
				$conf_led_color			\
			]	\
		]
	}

	public method set_config {state} {
		if {[catch {
			array set connection_port [lindex $state 0]
			array set connection_pin [lindex $state 1]
			if {[string length [lindex $state 2]]} {
				wm geometry $win [lindex $state 2]
			}
			
			$canvas_widget.usr_note delete 0
			$canvas_widget.usr_note insert 0 [lindex $state 3]
			
			set conf_led_color [lindex $state 4]
		
			for {set i 0} {$i < 8} {incr i} {
				set pin $connection_pin($i)
				if {$pin != {-}} {
					set pin	[expr {7 - $pin}]
				}
				set idx [lsearch -ascii -exact			\
					[$canvas_widget.cb_b$i cget -values]	\
					$pin					\
				]
				if {$idx == -1} {
					set idx 0
				}
				$canvas_widget.cb_b$i setvalue @$idx


				set idx [lsearch -ascii -exact			\
					[$canvas_widget.cb_p$i cget -values]	\
					$connection_port($i)			\
				]
				if {$idx == -1} {
					set idx 0
				}
				$canvas_widget.cb_p$i setvalue @$idx
			}

			new_state [$project pale_get_true_state]
			update

		}]} then {
			puts "Unable to load config for $class_name"
			return 0
		} else {
			clear_modified
			return 1
		}
	}
	
	public method reset {} {
		new_state [$project pale_get_true_state]
	}
	# ------------------------------------------------------

	public method config_menu_special {} {
		set ::${class_name}::color $conf_led_color
	}

	public method create_config_menu_special {} {
		foreach item {	Red	Orange	Yellow	Green	Blue	Purple	} \
			color {	#DD0000	#DD8800	#DDDD00	#00DD00	#0000DD	#8800DD	} \
		{
			$conf_menu.color entryconfigure $item -foreground $color
		}
	}
	
	public method color_changed {} {
		set conf_led_color ${::LedDisplay::color}
	
		if {$drawing_on} {
			new_state [$project pale_get_true_state]
		}
		set_modified
	}

	public method show_help_special {text_widget} {
# 		$text_widget insert insert "How to use this tool:\n"
# 		$text_widget tag add tag_bold insert-1l insert
# 		$text_widget insert insert "   1) Connect programmer to the computer\n"
# 		$text_widget tag add tag_bold insert-1l {insert-1l+6c}
# 		$text_widget insert insert "   2) Select port in section \"Programmer\"\n"
# 		$text_widget tag add tag_bold insert-1l {insert-1l+6c}
# 		$text_widget insert insert "   3) Click on button \"Autodetect device\"\n"
# 		$text_widget tag add tag_bold insert-1l {insert-1l+6c}
# 		$text_widget insert insert "   4) If programmer is operational you will see name of connected MCU in section \"Connected MCU\"\n"
# 		$text_widget tag add tag_bold insert-1l {insert-1l+6c}
# 		$text_widget insert insert "   5) Now you can perform arbitrary desired operation\n"
# 		$text_widget tag add tag_bold insert-1l {insert-1l+6c}
# 		$text_widget insert insert "\nMost useful functions are upload flash memory in section \"Flash memory (remember than flash must be first erased before programming). And \"Reprogram\" in section Special functions\", it will erase and reprogram all on-chip memory."
	}

	public method close_window_special {} {
	}
}
