#!/usr/bin/tclsh
# Part of MCU 8051 IDE ( http://mcu8051ide.sf.net )

############################################################################
#    Copyright (C) 2007-2009 by Martin Ošmera                              #
#    martin.osmera@gmail.com                                               #
#                                                                          #
#    This program is free software; you can redistribute it and#or modify  #
#    it under the terms of the GNU General Public License as published by  #
#    the Free Software Foundation; either version 2 of the License, or     #
#    (at your option) any later version.                                   #
#                                                                          #
#    This program is distributed in the hope that it will be useful,       #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#    GNU General Public License for more details.                          #
#                                                                          #
#    You should have received a copy of the GNU General Public License     #
#    along with this program; if not, write to the                         #
#    Free Software Foundation, Inc.,                                       #
#    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             #
############################################################################


# --------------------------------------------------------------------------
# DESCRIPTION
#
# --------------------------------------------------------------------------

class LedMatrix {
	inherit VirtualHWComponent
	
	common cb_font	[font create -weight bold -size -10 -family {helvetica}]
	common COMPONENT_NAME	"LED matrix"
	common CLASS_NAME	"LedMatrix"
	common COMPONENT_ICON	{ledmatrix}

	# Configuration menu
	common CONFMENU {
		{cascade	{Dim interval}		4	"player_time"	.dim	false 1 {
			{radiobutton	"0"		{}
				::LedMatrix::dim_interval	0
				"dim_interval_changed"	-1
				"Set LED dim interval to 0 instruction cycles"}
			{radiobutton	"5"		{}
				::LedMatrix::dim_interval	5
				"dim_interval_changed"	-1
				"Set LED dim interval to 5 instruction cycles"}
			{radiobutton	"10"		{}
				::LedMatrix::dim_interval	10
				"dim_interval_changed"	-1
				"Set LED dim interval to 10 instruction cycles"}
			{radiobutton	"20"		{}
				::LedMatrix::dim_interval	20
				"dim_interval_changed"	-1
				"Set LED dim interval to 20 instruction cycles"}
			{radiobutton	"50"		{}
				::LedMatrix::dim_interval	50
				"dim_interval_changed"	-1
				"Set LED dim interval to 50 instruction cycles"}
			{radiobutton	"100"		{}
				::LedMatrix::dim_interval	100
				"dim_interval_changed"	-1
				"Set LED dim interval to 100 instruction cycles"}
			{radiobutton	"200"		{}
				::LedMatrix::dim_interval	200
				"dim_interval_changed"	-1
				"Set LED dim interval to 200 instruction cycles"}
			{radiobutton	"500"		{}
				::LedMatrix::dim_interval	500
				"dim_interval_changed"	-1
				"Set LED dim interval to 500 instruction cycles"}
			{radiobutton	"1000"		{}
				::LedMatrix::dim_interval	1000
				"dim_interval_changed"	-1
				"Set LED dim interval to 1000 instruction cycles"}
		}}
		{cascade	{Color}			0	"colorize"	.color	false 1 {
			{radiobutton	"Red"		{}
				::LedMatrix::color		{red}
				"color_changed"	0	""}
			{radiobutton	"Orange"	{}
				::LedMatrix::color		{orange}
				"color_changed"	0	""}
			{radiobutton	"Yellow"	{}
				::LedMatrix::color		{yellow}
				"color_changed"	0	""}
			{radiobutton	"Green"		{}
				::LedMatrix::color		{green}
				"color_changed"	0	""}
			{radiobutton	"Blue"		{}
				::LedMatrix::color		{blue}
				"color_changed"	0	""}
			{radiobutton	"Purple"	{}
				::LedMatrix::color		{purple}
				"color_changed"	0	""}
		}}
		{cascade	{Light up when}		0	"ledgreen"	.cond	false 1 {
			{radiobutton	"Row 0 & Column 0"	{}
				::LedMatrix::cond		{0 0}
				"cond_changed"	0
				"Light up LED when both wires are in low"}
			{radiobutton	"Row 0 & Column 1"	{}
				::LedMatrix::cond		{0 1}
				"cond_changed"	0
				"Light up LED when row wire is in low and column wire is in high"}
			{radiobutton	"Row 1 & Column 0"	{}
				::LedMatrix::cond		{1 0}
				"cond_changed"	0
				"Light up LED when row wire is in high and column wire is in low"}
			{radiobutton	"Row 1 & Column 1"	{}
				::LedMatrix::cond		{1 1}
				"cond_changed"	0
				"Light up LED when both wires are in high"}
		}}
		{separator}
		{command	{Dim all}		{}	0	"dim_all"	{ledgray}
			"Dim all leds"}
		{command	{Show help}		{}	5	"show_help"	{help}
			"Show brief help"}
		{separator}
		{command	{Save configuration}	{}	0	"save_as"	{filesave}
			"Save configuration into a file"}
		{command	{Load configuration}	{}	0	"load_from"	{fileopen}
			"Load configuration from a file"}
	}

	private variable conf_dim_interval	50
	private variable conf_led_color		{red}
	private variable conf_led_cond		{0 0}

	private variable leds
	private variable col
	private variable row
	private variable prev_state
	private variable connection_port
	private variable connection_pin

	constructor {_project} {
		set component_name	$COMPONENT_NAME
		set class_name		$CLASS_NAME
		set component_icon	$COMPONENT_ICON
		
		set project $_project
		$project pale_register_output_device $this
		$project pale_set_modified

		array set connection_port {
			R0 - R1 - R2 - R3 - R4 - R5 - R6 - R7 -
			C0 - C1 - C2 - C3 - C4 - C5 - C6 - C7 -
		}
		array set connection_pin {
			R0 - R1 - R2 - R3 - R4 - R5 - R6 - R7 -
			C0 - C1 - C2 - C3 - C4 - C5 - C6 - C7 -
		}
		
		for {set j 0} {$j < 8} {incr j} {
			for {set i 0} {$i < 8} {incr i} {
				set prev_state($j,$i) 0
			}
		}

		create_gui
		mcu_changed
		on_off [$project pale_is_enabled]

		for {set i 0} {$i < 8} {incr i} {
			$canvas_widget.cb_r_b$i setvalue first
			$canvas_widget.cb_r_p$i setvalue first
			
			$canvas_widget.cb_c_b$i setvalue first
			$canvas_widget.cb_c_p$i setvalue first
		}
	}

	destructor {
		$project pale_unregister_output_device $this
		destroy $win
	}

	public method reconnect {col_or_row i} {
		if {$col_or_row == {R}} {
			set connection_port(R$i) [$canvas_widget.cb_r_p$i get]
			set connection_pin(R$i)	[$canvas_widget.cb_r_b$i get]
			if {$connection_pin(R$i) != {-}} {
				set connection_pin(R$i)	[expr {7 - $connection_pin(R$i)}]
			}
		} {
			set connection_port(C$i) [$canvas_widget.cb_c_p$i get]
			set connection_pin(C$i)	[$canvas_widget.cb_c_b$i get]
			if {$connection_pin(C$i) != {-}} {
				set connection_pin(C$i)	[expr {7 - $connection_pin(C$i)}]
			}
		}

		if {$drawing_on} {
			new_state [$project pale_get_true_state]
		}
		set_modified
	}

	private method create_gui {} {
		set win [toplevel .ledmatrix$count -class $component_name]

		set canvas_widget [canvas $win.canvas	\
			-bg white -width 0 -height 0	\
			-highlightthickness 0		\
		]

		set led_sep 15
		set sep 25
		set y_0 100
		set x_0 85
		set y_1 [expr {$y_0 + 10}]
		set x_1 [expr {$x_0 + 35}]

		set cb_x $x_0
		set cb_y [expr {$y_0 - 80}]

		set x $x_1
		set y $y_1
		incr y -$led_sep
		for {set i 0} {$i < 8} {incr i} {
			set fin [expr {$cb_y + 30 + abs($x - $cb_x)}]
			set col($i) [$canvas_widget create line	\
				$cb_x [expr {$cb_y + 20}]	\
				$cb_x [expr {$cb_y + 30}]	\
				$x $fin $x [expr {$y - 8}]	\
				-fill {#000000} -width 1	\
			]
			lappend col($i) [$canvas_widget create text $x $y	\
				-text $i -font $cb_font -fill {#000000}	\
			]
			incr x $led_sep
			incr cb_x $sep
		}

		
		set cb_x $x_0
		set cb_y $y_0
		incr cb_y -$sep
		incr cb_x -55

		set x $x_1
		set y $y_1
		incr x -$led_sep
		for {set i 0} {$i < 8} {incr i} {
			set fin [expr {$cb_x + 30 + abs($y - $cb_y)}]
			set row($i) [$canvas_widget create line		\
				$cb_x $cb_y [expr {$cb_x + 30}] $cb_y	\
				$fin $y [expr {$x - 8}] $y		\
				-fill {#000000} -width 1		\
			]
			lappend row($i) [$canvas_widget create text $x $y	\
				-font $cb_font -fill {#000000}		\
				-text [lindex {A B C D E F G H} $i]	\
			]
			
			incr y $led_sep
			incr cb_y $sep
		}
		
		set y $y_1
		set x $x_1
		for {set j 0} {$j < 8} {incr j} {
			for {set i 0} {$i < 8} {incr i} {
				set leds($j,$i) [$canvas_widget create image $x $y	\
					-image ::ICONS::16::ledgray			\
				]
				incr x $led_sep
			}
			incr y $led_sep
			set x $x_1
		}
		

		set x $x_0
		set y $y_0
		incr y -$sep
		incr x -55
		for {set i 0} {$i < 8} {incr i} {
			$canvas_widget create window $x $y -anchor e	\
				-window [ComboBox $canvas_widget.cb_r_p$i	\
					-width 1 -font $cb_font		\
					-modifycmd "$this reconnect R $i"	\
					-entrybg white -editable 0	\
					-hottrack 1			\
				]
			$canvas_widget create window $x $y -anchor w	\
				-window [ComboBox $canvas_widget.cb_r_b$i	\
					-width 1 -font $cb_font		\
					-modifycmd "$this reconnect R $i"	\
					-values {- 0 1 2 3 4 5 6 7}	\
					-entrybg white -editable 0	\
					-hottrack 1			\
				]

			catch {
				bindtags $canvas_widget.cb_r_p$i.e	\
					[list $canvas_widget.cb_r_p$i.e Entry all .]
				bindtags $canvas_widget.cb_r_b$i.e	\
					[list $canvas_widget.cb_r_b$i.e Entry all .]
			}
			
			incr y $sep
		}

		set x $x_0
		set cb_p_y [expr {$y_0 - 85}]
		set cb_b_y [expr {$y_0 - 65}]
		for {set i 0} {$i < 8} {incr i} {
			$canvas_widget create window $x $cb_p_y -anchor center	\
				-window [ComboBox $canvas_widget.cb_c_p$i	\
					-width 1 -font $cb_font		\
					-modifycmd "$this reconnect C $i"	\
					-entrybg white -editable 0	\
					-hottrack 1			\
				]
			$canvas_widget create window $x $cb_b_y -anchor center	\
				-window [ComboBox $canvas_widget.cb_c_b$i	\
					-width 1 -font $cb_font		\
					-modifycmd "$this reconnect C $i"	\
					-values {- 0 1 2 3 4 5 6 7}	\
					-entrybg white -editable 0	\
					-hottrack 1			\
				]

			catch {
				bindtags $canvas_widget.cb_c_p$i.e	\
					[list $canvas_widget.cb_c_p$i.e Entry all .]
				bindtags $canvas_widget.cb_c_b$i.e	\
					[list $canvas_widget.cb_c_b$i.e Entry all .]
			}

			incr x $sep
		}
 
		set x $x_0
		set y $y_0
		incr y -$sep
		incr x -55

		$canvas_widget create text 70 $cb_p_y	\
			-text {PORT}			\
			-font $cb_font			\
			-anchor e
		$canvas_widget create text 70 $cb_b_y	\
			-text {BIT}			\
			-font $cb_font			\
			-anchor e
		$canvas_widget create text [expr {$x_0 - 53}] [expr {$y_0 - 2*$sep}]	\
			-text {PORT}			\
			-font $cb_font			\
			-anchor ne
		$canvas_widget create text [expr {$x_0 - 50}] [expr {$y_0 - 2*$sep}]	\
			-text {BIT}			\
			-font $cb_font			\
			-anchor nw

		$canvas_widget create text 5 273	\
			-text {Note}			\
			-font $cb_font			\
			-anchor w
		$canvas_widget create window 40 273		\
			-window [entry $canvas_widget.usr_note	\
				-bg white -validate all		\
				-vcmd "$this set_modified"	\
			]	\
			-width 230 -anchor w
		bindtags $canvas_widget.usr_note	\
			[list $canvas_widget.usr_note Entry $win all .]

		
		set start_stop_button [Button $canvas_widget.start_stop_button	\
			-bd 1 -command "$this on_off_button_press"		\
			-padx 0 -pady 0 -highlightthickness 0 -width 3		\
			-helptext "Turn HW simulation on/off"			\
		]
		setStatusTip -widget $start_stop_button -text "Turn HW simulation on/off"
		bind $start_stop_button <Button-3> "$this on_off_button_press; break"
		$canvas_widget create window 2 20 -window $start_stop_button -anchor sw
		
		set conf_button [button $canvas_widget.conf_but	\
			-image ::ICONS::16::configure -bd 0	\
			-command "$this config_menu"		\
			-relief flat -overrelief raised		\
			-bg white -activebackground white	\
			-highlightthickness 0			\
		]
		setStatusTip -widget $conf_button -text "Configure"
		$canvas_widget create window 2 20 -window $conf_button -anchor nw
		
		pack $canvas_widget -fill both -expand 1
		
		# Set window parameters
		wm geometry $win =275x285
		wm iconphoto $win ::ICONS::16::$component_icon
		wm title $win "$component_name - [string trim $project {:}] - MCU 8051 IDE"
		wm resizable $win 0 0
		wm protocol $win WM_DELETE_WINDOW "$this close_window"
		
		bindtags $win [list $win Toplevel all .]
	}

	private method which_port_pin {col_or_row i} {
		return [list $connection_port(${col_or_row}${i}) $connection_pin(${col_or_row}${i})]
	}

	public method on_off_button_press {} {
		$project pale_all_on_off
	}

	# ------------------------------------------------------------------
	# VIRTUAL HW COMMON INTERFACE
	# ------------------------------------------------------------------

	public method mcu_changed {} {
		set avaliable_ports [concat - [$project pale_get_avaliable_ports]]
		
		for {set i 0} {$i < 8} {incr i} {
			$canvas_widget.cb_r_p$i configure -values $avaliable_ports
			$canvas_widget.cb_c_p$i configure -values $avaliable_ports

			if {[lsearch -ascii -exact $avaliable_ports $connection_port(R$i)] == -1} {
				$canvas_widget.cb_r_p$i setvalue first
				set connection_port(R$i) {-}
			}
			if {[lsearch -ascii -exact $avaliable_ports $connection_port(C$i)] == -1} {
				$canvas_widget.cb_c_p$i setvalue first
				set connection_port(C$i) {-}
			}
		}
	}

	public method on_off {state} {
		set drawing_on $state
		
		if {$state} {
			$start_stop_button configure	-text {ON}			\
				-bg {#AAFFAA}		-activebackground {#DDFFDD}	\
				-fg {#000000}		-activeforeground {#00FF00}

			new_state [$project pale_get_true_state]

		} else {
			$start_stop_button configure	-text {OFF}			\
				-bg {#FFAAAA}		-activebackground {#FFDDDD}	\
				-fg {#000000}		-activeforeground {#FF0000}

			for {set j 0} {$j < 8} {incr j} {
				for {set i 0} {$i < 8} {incr i} {
					$canvas_widget itemconfigure $leds($j,$i)	\
						-image ::ICONS::16::ledgray
				}
			}
		}
	}

	public method new_state {state} {
		for {set i 0} {$i < 8} {incr i} {
			set pp [which_port_pin C $i]
			if {[lindex $pp 0] == {-} || [lindex $pp 1] == {-}} {
				foreach item $col($i) {
					$canvas_widget itemconfigure $item -fill {#000000}
				}
				continue
			}

			switch -- [lindex $state $pp] {
				{0} {
					set wire_color		{#00FF00}
				}
				{1} {
					set wire_color		{#FF0000}
				}
				{=} {
					set wire_color		{#FF00AA}
				}
				{} {
					set wire_color		{#000000}
				}
				{?} {
					set wire_color		{#888888}
				}
				default {
					set wire_color		{#FF8800}
				}
			}

			foreach item $col($i) {
				$canvas_widget itemconfigure $item -fill $wire_color
			}
		}
		for {set i 0} {$i < 8} {incr i} {
			set pp [which_port_pin R $i]
			if {[lindex $pp 0] == {-} || [lindex $pp 1] == {-}} {
				foreach item $row($i) {
					$canvas_widget itemconfigure $item -fill {#000000}
				}
				continue
			}

			switch -- [lindex $state $pp] {
				{0} {
					set wire_color		{#00FF00}
				}
				{1} {
					set wire_color		{#FF0000}
				}
				{=} {
					set wire_color		{#FF00AA}
				}
				{} {
					set wire_color		{#000000}
				}
				{?} {
					set wire_color		{#888888}
				}
				default {
					set wire_color		{#FF8800}
				}
			}

			foreach item $row($i) {
				$canvas_widget itemconfigure $item -fill $wire_color
			}
		}

		for {set j 0} {$j < 8} {incr j} {
			set pp [which_port_pin R $j]
			if {[lindex $pp 0] == {-} || [lindex $pp 1] == {-}} {
				for {set i 0} {$i < 8} {incr i} {
					$canvas_widget itemconfigure	\
						$leds($j,$i)		\
						-image ::ICONS::16::ledgray
				}
				continue
			}

			set row_state [lindex $state $pp]
			switch -- $row_state {
				{0} {
					if {[lindex $conf_led_cond 0]} {
						set row_state 0
					} {
						set row_state 2
					}
				}
				{1} {
					if {[lindex $conf_led_cond 0]} {
						set row_state 2
					} {
						set row_state 0
					}
				}
				{?} {
					set row_state 0
				}
				{=} {
					set row_state 0
				}
				{} {
					set row_state 0
				}
				default {
					set row_state 1
				}
			}

			for {set i 0} {$i < 8} {incr i} {
				set pp [which_port_pin C $i]
				if {[lindex $pp 0] == {-} || [lindex $pp 1] == {-}} {
					$canvas_widget itemconfigure $leds($j,$i) -image ::ICONS::16::ledgray
					continue
				}

				switch -- [lindex $state $pp] {
					{0} {
						if {[lindex $conf_led_cond 1]} {
							set image ledgray
						} {
							set image shine
						}
					}
					{1} {
						if {[lindex $conf_led_cond 1]} {
							set image shine
						} {
							set image ledgray
						}
					}
					{?} {
						set image ledgray
					}
					{=} {
						set image ledgray
					}
					{} {
						set image ledgray
					}
					default {
						switch -- $row_state {
							{0} {
								set image ledgray
							}
							{1} {
								set image ledgray
							}
							{2} {
								set image ledgray${conf_led_color}
							}
						}
					}
				}

				if {$image == {shine}} {
					switch -- $row_state {
						{0} {
							set image ledgray
						}
						{1} {
							set image ledgray${conf_led_color}
						}
						{2} {
							set image led${conf_led_color}
						}
					}
				}

				if {$image == {ledgray}} {
					if {$prev_state($j,$i)} {
						incr prev_state($j,$i) -1
						if {$prev_state($j,$i)} {
							set image led${conf_led_color}2
						}
					}
				} elseif {$image == "led${conf_led_color}"} {
					set prev_state($j,$i) $conf_dim_interval
				}

				$canvas_widget itemconfigure $leds($j,$i)	\
					-image ::ICONS::16::$image
			}
		}
	}

	public method withdraw_window {} {
		wm withdraw $win
	}

	public method get_config {} {
		return [list		\
			$class_name	\
			[list		\
				[array get connection_port]	\
				[array get connection_pin]	\
				[wm geometry $win]		\
				[$canvas_widget.usr_note get]	\
				$conf_led_color			\
				$conf_dim_interval		\
				$conf_led_cond			\
			]	\
		]
	}

	public method set_config {state} {
		if {[catch {
			array set connection_port [lindex $state 0]
			array set connection_pin [lindex $state 1]
			if {[string length [lindex $state 2]]} {
				wm geometry $win [lindex $state 2]
			}
			
			$canvas_widget.usr_note delete 0
			$canvas_widget.usr_note insert 0 [lindex $state 3]
			
			set conf_led_color	[lindex $state 4]
			set conf_dim_interval	[lindex $state 5]
			set conf_led_cond	[lindex $state 6]

			foreach foo {R C} bar {r c} {
				for {set i 0} {$i < 8} {incr i} {
					set pin $connection_pin(${foo}${i})
					if {$pin != {-}} {
						set pin	[expr {7 - $pin}]
					}
					set idx [lsearch -ascii -exact				\
						[$canvas_widget.cb_${bar}_b$i cget -values]	\
						$pin						\
					]
					if {$idx == -1} {
						set idx 0
					}
					$canvas_widget.cb_${bar}_b$i setvalue @$idx

					set idx [lsearch -ascii -exact				\
						[$canvas_widget.cb_${bar}_p$i cget -values]	\
						$connection_port(${foo}$i)			\
					]
					if {$idx == -1} {
						set idx 0
					}
					$canvas_widget.cb_${bar}_p$i setvalue @$idx
				}
			}

			new_state [$project pale_get_true_state]
			update

		}]} then {
			puts "Unable to load config for $class_name"
			return 0
		} else {
			clear_modified
			return 1
		}
	}

	public method reset {} {
		dim_all
		new_state [$project pale_get_true_state]
	}

	# ------------------------------------------------------

	public method config_menu_special {} {
		set ::${class_name}::dim_interval $conf_dim_interval
		set ::${class_name}::color $conf_led_color
		set ::${class_name}::cond $conf_led_cond
	}
	
	public method create_config_menu_special {} {
		foreach item {	Red	Orange	Yellow	Green	Blue	Purple	} \
			color {	#DD0000	#DD8800	#DDDD00	#00DD00	#0000DD	#8800DD	} \
		{
			$conf_menu.color entryconfigure $item -foreground $color
		}
	}

	public method dim_interval_changed {} {
		set conf_dim_interval ${::LedMatrix::dim_interval}
		set_modified
	}
	
	public method color_changed {} {
		set conf_led_color ${::LedMatrix::color}
	
		if {$drawing_on} {
			new_state [$project pale_get_true_state]
		}
		set_modified
	}

	public method cond_changed {} {
		set conf_led_cond ${::LedMatrix::cond}
	
		if {$drawing_on} {
			new_state [$project pale_get_true_state]
		}
		set_modified
	}
	
	public method dim_all {} {
		for {set i 0} {$i < 8} {incr i} {
			for {set j 0} {$j < 8} {incr j} {
				if {$prev_state($j,$i)} {
					set prev_state($j,$i) 0
					$canvas_widget itemconfigure $leds($j,$i)	\
						-image ::ICONS::16::ledgray
				}
			}
		}
	}

	public method show_help_special {text_widget} {
# 		$text_widget insert insert "How to use this tool:\n"
# 		$text_widget tag add tag_bold insert-1l insert
# 		$text_widget insert insert "   1) Connect programmer to the computer\n"
# 		$text_widget tag add tag_bold insert-1l {insert-1l+6c}
# 		$text_widget insert insert "   2) Select port in section \"Programmer\"\n"
# 		$text_widget tag add tag_bold insert-1l {insert-1l+6c}
# 		$text_widget insert insert "   3) Click on button \"Autodetect device\"\n"
# 		$text_widget tag add tag_bold insert-1l {insert-1l+6c}
# 		$text_widget insert insert "   4) If programmer is operational you will see name of connected MCU in section \"Connected MCU\"\n"
# 		$text_widget tag add tag_bold insert-1l {insert-1l+6c}
# 		$text_widget insert insert "   5) Now you can perform arbitrary desired operation\n"
# 		$text_widget tag add tag_bold insert-1l {insert-1l+6c}
# 		$text_widget insert insert "\nMost useful functions are upload flash memory in section \"Flash memory (remember than flash must be first erased before programming). And \"Reprogram\" in section Special functions\", it will erase and reprogram all on-chip memory."
	}

	public method close_window_special {} {
	}
}
