#!/usr/bin/tclsh
# Part of MCU 8051 IDE ( http://mcu8051ide.sf.net )

############################################################################
#    Copyright (C) 2007-2009 by Martin Ošmera                              #
#    martin.osmera@gmail.com                                               #
#                                                                          #
#    This program is free software; you can redistribute it and#or modify  #
#    it under the terms of the GNU General Public License as published by  #
#    the Free Software Foundation; either version 2 of the License, or     #
#    (at your option) any later version.                                   #
#                                                                          #
#    This program is distributed in the hope that it will be useful,       #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#    GNU General Public License for more details.                          #
#                                                                          #
#    You should have received a copy of the GNU General Public License     #
#    along with this program; if not, write to the                         #
#    Free Software Foundation, Inc.,                                       #
#    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             #
############################################################################


# --------------------------------------------------------------------------
# DESCRIPTION
#
# --------------------------------------------------------------------------

class MultiplexedLedDisplay {
	inherit VirtualHWComponent

	#
	common cb_font		[font create	\
		-weight bold -size -10		\
		-family {helvetica}		\
	]
	common COMPONENT_NAME	"Multiplexed LED display"	;#
	common CLASS_NAME	"MultiplexedLedDisplay"		;#
	common COMPONENT_ICON	{mleddisplay}			;#

	# red orange yellow   green blue purple
	common COLORS {
		{#AA5555 #FF0000 #FF5555}
		{#AAAA55 #FF8800 #FFCC55}
		{#AAAA55 #FFFF00 #FFFF55}

		{#55AA55 #00FF00 #55FF55}
		{#5555AA #0000FF #5555FF}
		{#AA55AA #8800FF #CC55FF}
	}

	# Configuration menu
	common CONFMENU {
		{cascade	{Dim interval}		4	"player_time"	.dim	false 1 {
			{radiobutton	"0"		{}
				::MultiplexedLedDisplay::dim_interval	0
				"dim_interval_changed"	-1
				"Set LED dim interval to 0 instruction cycles"}
			{radiobutton	"5"		{}
				::MultiplexedLedDisplay::dim_interval	5
				"dim_interval_changed"	-1
				"Set LED dim interval to 5 instruction cycles"}
			{radiobutton	"10"		{}
				::MultiplexedLedDisplay::dim_interval	10
				"dim_interval_changed"	-1
				"Set LED dim interval to 10 instruction cycles"}
			{radiobutton	"20"		{}
				::MultiplexedLedDisplay::dim_interval	20
				"dim_interval_changed"	-1
				"Set LED dim interval to 20 instruction cycles"}
			{radiobutton	"50"		{}
				::MultiplexedLedDisplay::dim_interval	50
				"dim_interval_changed"	-1
				"Set LED dim interval to 50 instruction cycles"}
			{radiobutton	"100"		{}
				::MultiplexedLedDisplay::dim_interval	100
				"dim_interval_changed"	-1
				"Set LED dim interval to 100 instruction cycles"}
			{radiobutton	"200"		{}
				::MultiplexedLedDisplay::dim_interval	200
				"dim_interval_changed"	-1
				"Set LED dim interval to 200 instruction cycles"}
			{radiobutton	"500"		{}
				::MultiplexedLedDisplay::dim_interval	500
				"dim_interval_changed"	-1
				"Set LED dim interval to 500 instruction cycles"}
			{radiobutton	"1000"		{}
				::MultiplexedLedDisplay::dim_interval	1000
				"dim_interval_changed"	-1
				"Set LED dim interval to 1000 instruction cycles"}
		}}
		{cascade	{Color}			0	"colorize"	.color	false 1 {
			{radiobutton	"Red"		{}
				::MultiplexedLedDisplay::color		{red}
				"color_changed"	0	""}
			{radiobutton	"Orange"	{}
				::MultiplexedLedDisplay::color		{orange}
				"color_changed"	0	""}
			{radiobutton	"Yellow"	{}
				::MultiplexedLedDisplay::color		{yellow}
				"color_changed"	0	""}
			{radiobutton	"Green"		{}
				::MultiplexedLedDisplay::color		{green}
				"color_changed"	0	""}
			{radiobutton	"Blue"		{}
				::MultiplexedLedDisplay::color		{blue}
				"color_changed"	0	""}
			{radiobutton	"Purple"	{}
				::MultiplexedLedDisplay::color		{purple}
				"color_changed"	0	""}
		}}
		{separator}
		{command	{Dim all}		{}	0	"dim_all"	{ledgray}
			"Dim all leds"}
		{command	{Show help}		{}	5	"show_help"	{help}
			"Show brief help"}
		{separator}
		{command	{Save configuration}	{}	0	"save_as"	{filesave}
			"Save configuration into a file"}
		{command	{Load configuration}	{}	0	"load_from"	{fileopen}
			"Load configuration from a file"}
	}

	private variable conf_dim_interval	50	;#
	private variable conf_led_color		{red}	;#

	private variable leds				;#
	private variable wires				;#
	private variable connection_port		;#
	private variable connection_pin			;#
	private variable prev_state			;#

	private variable t_state			;#

	## Object constructor
	 # @parm
	constructor {_project} {
		set component_name	$COMPONENT_NAME
		set class_name		$CLASS_NAME
		set component_icon	$COMPONENT_ICON
	
		set project $_project
		
		$project pale_register_output_device $this
		$project pale_set_modified

		array set connection_port {
			0 - 1 - 2 - 3 - 4 - 5 - 6 - 7 -
			T0 - T1 - T2 - T3 -
		}
		array set connection_pin {
			0 - 1 - 2 - 3 - 4 - 5 - 6 - 7 -
			T0 - T1 - T2 - T3 -
		}
		
		for {set j 0} {$j < 4} {incr j} {
			for {set i 0} {$i < 8} {incr i} {
				set prev_state($j,$i) 0
			}
		}

		create_gui
		mcu_changed
		on_off [$project pale_is_enabled]

		for {set i 0} {$i < 8} {incr i} {
			$canvas_widget.cb_b$i setvalue first
			$canvas_widget.cb_p$i setvalue first
		}
		for {set i 0} {$i < 4} {incr i} {
			$canvas_widget.cb_t_p$i setvalue first
			$canvas_widget.cb_t_b$i setvalue first
		}
	}

	## Object destructor
	destructor {
		$project pale_unregister_output_device $this
		destroy $win
	}

	private method create_gui {} {
		set win [toplevel .multiplexedleddisplay$count -class {Mult. LED Display}]

		set canvas_widget [canvas $win.canvas	\
			-bg white -width 0 -height 0	\
			-highlightthickness 0		\
		]

		set x 90
		set y 15
		draw_connections $x $y
		for {set i 0} {$i < 4} {incr i} {
			draw_8_segment $x $y [expr {3 - $i}]
			draw_wires $x $y [lindex {0 1 1 0} $i]

			incr x 80
		}

		set tx_p_y 150
		set cb_p_y 168
		set cb_b_y 188
		set x 110
		for {set i 0} {$i < 8} {incr i} {
			$canvas_widget create window $x $cb_p_y -anchor center	\
				-window [ComboBox $canvas_widget.cb_p$i	\
					-width 1 -font $cb_font		\
					-modifycmd "$this reconnect S $i"	\
					-entrybg white -editable 0	\
					-hottrack 1			\
				]
			$canvas_widget create window $x $cb_b_y -anchor center	\
				-window [ComboBox $canvas_widget.cb_b$i	\
					-width 1 -font $cb_font		\
					-modifycmd "$this reconnect S $i"	\
					-values {- 0 1 2 3 4 5 6 7}	\
					-entrybg white -editable 0	\
					-hottrack 1			\
				]

			catch {
				bindtags $canvas_widget.cb_p$i.e	\
					[list $canvas_widget.cb_p$i.e Entry all .]
				bindtags $canvas_widget.cb_b$i.e	\
					[list $canvas_widget.cb_b$i.e Entry all .]
			}
			
			lappend wires($i) [$canvas_widget create text	\
				[expr {$x - 3}] $tx_p_y			\
				-text [lindex {A B C D E F G P} $i]	\
				-font $cb_font -fill #000000 -anchor e	\
			]
			
			lappend wires($i) [$canvas_widget create line	\
				$x $cb_p_y $x [expr {100 + $y + 4*$i}]	\
				-width 1 -fill #000000	\
			]

			if {[lindex {0 0 1 1 1 1 0 1} $i]} {
				lappend wires($i) [$canvas_widget create oval		\
					[expr {$x - 2}] [expr {100 + $y + 4*$i - 2}]	\
					[expr {$x + 2}] [expr {100 + $y + 4*$i + 2}]		\
					-width 0 -fill #000000				\
				]
			}

			incr x 40
		}

		set x 90
		set y 15
		foreach coords {
			{244 122	248 126}
			{76 98		80 102}
			{72 102		76 106}
		} index {
			6
			0
			1
		} {
			set coordinates [list]
			set len [llength $coords]

			for {set m 0; set n 1} {$n < $len} {incr m 2; incr n 2} {
				lappend coordinates	\
					[expr {[lindex $coords $m] + $x}]
				lappend coordinates	\
					[expr {[lindex $coords $n] + $y}]
			}
			
			lappend wires($index) [$canvas_widget create oval	\
				$coordinates -width 0 -fill #000000		\
			]
		}
		
		set cb_x 30
		set cb_y 55
		set tr_x 55
		set tr_y 45
		
		set txA_x 50
		set txA_y 33
		set txB_x 105
		set txB_y 30
		for {set i 0} {$i < 4} {incr i} {
			draw_transistor $tr_x $tr_y $i
			
			$canvas_widget create window $cb_x $cb_y -anchor e	\
				-window [ComboBox $canvas_widget.cb_t_p$i	\
					-width 1 -font $cb_font		\
					-modifycmd "$this reconnect T $i"	\
					-entrybg white -editable 0	\
					-hottrack 1			\
				]
			$canvas_widget create window $cb_x $cb_y -anchor w	\
				-window [ComboBox $canvas_widget.cb_t_b$i	\
					-width 1 -font $cb_font		\
					-modifycmd "$this reconnect T $i"	\
					-values {- 0 1 2 3 4 5 6 7}	\
					-entrybg white -editable 0	\
					-hottrack 1			\
				]

			catch {
				bindtags $canvas_widget.cb_t_p$i.e	\
					[list $canvas_widget.cb_t_p$i.e Entry all .]
				bindtags $canvas_widget.cb_t_b$i.e	\
					[list $canvas_widget.cb_t_b$i.e Entry all .]
			}

			set x_foo [expr {$tr_x + 18 + 4 + 4*$i}]
			set x_bar [expr {135 + (3 - $i)*80}]
			set wires(T$i) [$canvas_widget create line	\
				[expr {$tr_x + 18}] $tr_y	\
				$x_foo $tr_y	\
				$x_foo [expr {2 + $i*4}]	\
				$x_bar [expr {2 + $i*4}]	\
				$x_bar 19	\
				-width 1 -fill #000000	\
			]
			lappend wires(T$i) [$canvas_widget create line	\
				[expr {$x_bar - 38}] 30	\
				[expr {$x_bar - 38}] 19	\
				[expr {$x_bar + 38}] 19	\
				[expr {$x_bar + 38}] 30	\
				-width 2 -fill #000000	\
			]

			$canvas_widget create text $txA_x $txA_y	\
				-text $i -font $cb_font
			$canvas_widget create text $txB_x $txB_y	\
				-text [expr {3 - $i}] -font $cb_font

			incr tr_y 40
			incr cb_y 40
			incr txA_y 40
			incr txB_x 80
		}
		
		set start_stop_button [Button $canvas_widget.start_stop_button	\
			-bd 1 -command "$this on_off_button_press"		\
			-padx 0 -pady 0 -highlightthickness 0 -width 3		\
			-helptext "Turn HW simulation on/off"			\
		]
		setStatusTip -widget $start_stop_button -text "Turn HW simulation on/off"
		bind $start_stop_button <Button-3> "$this on_off_button_press; break"
		$canvas_widget create window 22 2 -window $start_stop_button -anchor nw
		
		set conf_button [button $canvas_widget.conf_but	\
			-image ::ICONS::16::configure -bd 0	\
			-command "$this config_menu"		\
			-relief flat -overrelief raised		\
			-bg white -activebackground white	\
			-highlightthickness 0			\
		]
		setStatusTip -widget $conf_button -text "Configure"
		$canvas_widget create window 20 2 -window $conf_button -anchor ne


		$canvas_widget create text 40 210	\
			-text {Note}			\
			-font $cb_font			\
			-anchor e
		$canvas_widget create window 40 210		\
			-window [entry $canvas_widget.usr_note	\
				-bg white -validate all		\
				-vcmd "$this set_modified"	\
			]	\
			-width 380 -anchor w
		bindtags $canvas_widget.usr_note	\
			[list $canvas_widget.usr_note Entry $win all .]


		pack $canvas_widget -fill both -expand 1
		
		# Set window parameters
		wm geometry $win =425x225
		wm iconphoto $win ::ICONS::16::$component_icon
		wm title $win "$component_name - [string trim $project {:}] - MCU 8051 IDE"
		wm resizable $win 0 0
		wm protocol $win WM_DELETE_WINDOW "$this close_window"
		
		bindtags $win [list $win Toplevel all .]
	}

	public method reconnect {type i} {
		if {$type == {S}} {
			set connection_port($i) [$canvas_widget.cb_p$i get]
			set connection_pin($i)	[$canvas_widget.cb_b$i get]
			if {$connection_pin($i) != {-}} {
				set connection_pin($i)	[expr {7 - $connection_pin($i)}]
			}
		} {
			set connection_port(T$i) [$canvas_widget.cb_t_p$i get]
			set connection_pin(T$i)	[$canvas_widget.cb_t_b$i get]
			if {$connection_pin(T$i) != {-}} {
				set connection_pin(T$i)	[expr {7 - $connection_pin(T$i)}]
			}
		}

		if {$drawing_on} {
			new_state [$project pale_get_true_state]
		}
		set_modified
	}

	private method draw_transistor {x y index} {
		set coords {
			0 9	8 9	8 0	8 18	8 9
			9 9	18 0	9 9	11 11	11 15
			11 11	15 11	11 11	18 18	18 24
		}

		set coordinates [list]
		set len [llength $coords]

		for {set m 0; set n 1} {$n < $len} {incr m 2; incr n 2} {
			lappend coordinates	\
				[expr {[lindex $coords $m] + $x}]
			lappend coordinates	\
				[expr {[lindex $coords $n] + $y}]
		}
			
		$canvas_widget create line $coordinates	\
			-tags transistor -width 1 -fill #000000

		$canvas_widget create oval 			\
			[expr {$x + 15}] [expr {$y + 24}]	\
			[expr {$x + 21}] [expr {$y + 30}]	\
			-tags transistor -width 1 -outline #000000

		$canvas_widget create text 			\
			[expr {$x + 15}] [expr {$y + 27}]	\
			-tags transistor -font $::smallfont	\
			-fill #000000 -text {+} -anchor e
	}

	private method draw_connections {x y} {
		set coords {
			{
				20 100 319 100
			} {
				60 104	315 104
			} {
				70 108	311 108
			} {
				35 112	276 112
			} {
				10 116	251 116
			} {
				2 120	243 120
			} {
				6 124	260 124
			} {
				64 128	305 128
			}
		}

		for {set i 0} {$i < 8} {incr i} {
			set coordinates [list]
			set local_coords [lindex $coords $i]
			set len [llength $local_coords]

			for {set m 0; set n 1} {$n < $len} {incr m 2; incr n 2} {
				lappend coordinates	\
					[expr {[lindex $local_coords $m] + $x}]
				lappend coordinates	\
					[expr {[lindex $local_coords $n] + $y}]
			}

			set wires($i) [$canvas_widget create line	\
				$coordinates -width 1 -fill #000000	\
			]
		}
	}

	private method draw_wires {x y junction} {
		set coords {
			{
				46 10	46 8	78 8	78 100
			} {
				70 37	74 37	74 104
			} {
				64 72	70 72	70 108
			} {
				35 95	35 112
			} {
				14 70	10 70	10 116
			} {
				19 33	2 33	2 120
			} {
				25 53	6 53	6 124
			} {
				64 96	64 128
			}
		}

		for {set i 0} {$i < 8} {incr i} {
			set coordinates [list]
			set local_coords [lindex $coords $i]
			set len [llength $local_coords]

			for {set m 0; set n 1} {$n < $len} {incr m 2; incr n 2} {
				lappend coordinates	\
					[expr {[lindex $local_coords $m] + $x}]
				lappend coordinates	\
					[expr {[lindex $local_coords $n] + $y}]
			}

			lappend wires($i) [$canvas_widget create line	\
				$coordinates -width 1 -fill #000000	\
			]

			if {$junction} {
				set oval [expr {[lindex $coordinates end-1] - 2}]
				lappend oval [expr {[lindex $coordinates end] - 2}]
				lappend oval [expr {[lindex $coordinates end-1] + 2}]
				lappend oval [expr {[lindex $coordinates end] + 2}]
				lappend wires($i) [$canvas_widget create oval	\
					$oval -width 0 -fill #000000	\
				]
			}
		}
	}

	private method draw_8_segment {x y index} {
		set coords {
			{
				19 7	25 1	47 1	53 7	53 8
				47 14	25 14	19 8
			} {
				55 9	62 16	58 34	50 42	44 36
				49 15	55 9
			} {
				50 45	57 52	53 70	46 77	45 77
				39 71	44 51	50 45
			} {
				15 73	38 73	44 79	37 86	15 86
				9 80	9 79
			} {
				7 78	15 70	19 52	12 45	5 52
				1 72
			} {
				12 42	20 34	25 16	17 9	10 16
				6 36
			} {
				14 43	20 37	42 37	48 43	48 44
				42 50	20 50	14 44
			}
		}

		for {set i 0} {$i < 7} {incr i} {
			set coordinates [list]
			set local_coords [lindex $coords $i]
			set len [llength $local_coords]

			for {set m 0; set n 1} {$n < $len} {incr m 2; incr n 2} {
				lappend coordinates	\
					[expr {[lindex $local_coords $m] + $x + 10}]
				lappend coordinates	\
					[expr {[lindex $local_coords $n] + $y + 10}]
			}

			set leds($index,$i) [$canvas_widget create polygon	\
				$coordinates -width 0 -fill #555555	\
			]
		}
		set leds($index,7) [$canvas_widget create oval			\
			[expr {49 + $x + 10}] [expr {77 + $y + 10}]	\
			[expr {58 + $x + 10}] [expr {86 + $y + 10}]	\
			-width 0 -fill #555555				\
		]
		
		foreach coords {{35 7} {53 25} {48 62} {26 79} {10 61} {15 25} {31 43}} \
			text {A B C D E F G} \
		{
			$canvas_widget create text			\
				[expr {[lindex $coords 0] + $x + 10}]	\
				[expr {[lindex $coords 1] + $y + 10}]	\
				-text $text -fill {#FFFFFF}		\
				-font $::smallfont
		}
	}

	private method which_port_pin {type i} {
		if {$type == {S}} {
			return [list $connection_port($i) $connection_pin($i)]
		} {
			return [list $connection_port(T$i) $connection_pin(T$i)]
		}
	}

	public method on_off_button_press {} {
		$project pale_all_on_off
	}


	# ------------------------------------------------------------------
	# VIRTUAL HW COMMON INTERFACE
	# ------------------------------------------------------------------

	public method mcu_changed {} {
		set avaliable_ports [concat - [$project pale_get_avaliable_ports]]
		
		for {set i 0} {$i < 8} {incr i} {
			$canvas_widget.cb_p$i configure -values $avaliable_ports

			if {[lsearch -ascii -exact $avaliable_ports $connection_port($i)] == -1} {
				$canvas_widget.cb_p$i setvalue first
				set connection_port($i) {-}
			}
		}

		for {set i 0} {$i < 4} {incr i} {
			$canvas_widget.cb_t_p$i configure -values $avaliable_ports

			if {[lsearch -ascii -exact $avaliable_ports $connection_port(T$i)] == -1} {
				$canvas_widget.cb_t_p$i setvalue first
				set connection_port(T$i) {-}
			}
		}
	}
	
	public method on_off {state} {
		set drawing_on $state
		
		if {$state} {
			$start_stop_button configure	-text {ON}			\
				-bg {#AAFFAA}		-activebackground {#DDFFDD}	\
				-fg {#000000}		-activeforeground {#00FF00}

			new_state [$project pale_get_true_state]

		} else {
			$start_stop_button configure	-text {OFF}			\
				-bg {#FFAAAA}		-activebackground {#FFDDDD}	\
				-fg {#000000}		-activeforeground {#FF0000}

			for {set j 0} {$j < 4} {incr j} {
				for {set i 0} {$i < 8} {incr i} {
					$canvas_widget itemconfigure $leds($j,$i) -fill {#555555}
					foreach item $wires($i) {
						$canvas_widget itemconfigure $item -fill {#000000}
					}
				}
			}
		}
	}

	public method new_state {state} {	
		set color_idx [lsearch -ascii -exact		\
			{red orange yellow green blue purple}	\
			$conf_led_color				\
		]

		for {set i 0} {$i < 4} {incr i} {
			set pp [which_port_pin T $i]
			if {[lindex $pp 0] == {-} || [lindex $pp 1] == {-}} {
				set t_state($i) {}
			} {
				set t_state($i) [lindex $state $pp]
			}

			switch -- $t_state($i) {
				{0} {
					set t_state($i)	2
					set wire_color	{#FF0000}
				}
				{1} {
					set t_state($i)	0
					set wire_color	{#00FF00}
				}
				{=} {
					set t_state($i)	2
					set wire_color	{#00FF00}
				}
				{?} {
					set t_state($i)	0
					set wire_color	{#888888}
				}
				{} {
					set t_state($i)	0
					set wire_color	{#000000}
				}
				default {
					set t_state($i)	1
					set wire_color	{#FF8800}
				}
			}
			
			foreach item $wires(T$i) {
				$canvas_widget itemconfigure $item -fill $wire_color
			}
		}

		for {set i 0} {$i < 8} {incr i} {
			set pp [which_port_pin S $i]
			if {[lindex $pp 0] == {-} || [lindex $pp 1] == {-}} {
				for {set j 0} {$j < 4} {incr j} {
					$canvas_widget itemconfigure $leds($j,$i) -fill {#555555}
				}
				foreach item $wires($i) {
					$canvas_widget itemconfigure $item -fill {#000000}
				}
				continue
			}
			set state_pp [lindex $state $pp]

			switch -- $state_pp {
				{0} {
					set wire_color	{#00FF00}
				}
				{1} {
					set wire_color	{#FF0000}
				}
				{=} {
					set wire_color	{#FF00AA}
				}
				{} {
					set wire_color	{#000000}
				}
				{?} {
					set wire_color	{#888888}
				}
				default {
					set wire_color	{#FF8800}
				}
			}

			foreach item $wires($i) {
				$canvas_widget itemconfigure $item -fill $wire_color
			}

			for {set j 0} {$j < 4} {incr j} {
				switch -- $state_pp {
					{0} {
						switch -- $t_state($j) {
							{2} {
								set segment_color {2}
							}
							{1} {
								set segment_color {1}
							}
							{0} {
								set segment_color {0}
							}
						}
					}
					{1} {
						set segment_color {0}
					}
					{=} {
						set segment_color {0}
					}
					{} {
						set segment_color {0}
					}
					{?} {
						set segment_color {0}
					}
					default {
						switch -- $t_state($j) {
							{2} {
								set segment_color {1}
							}
							{1} {
								set segment_color {0}
							}
							{0} {
								set segment_color {0}
							}
						}
					}
				}

				set outline {#FFFFFF}
				if {$segment_color == {0}} {
					if {$prev_state($j,$i)} {
						incr prev_state($j,$i) -1
						if {$prev_state($j,$i)} {
							set segment_color {3}
						}
					}
				} elseif {$segment_color == {2}} {
					set prev_state($j,$i) $conf_dim_interval
					set outline [lindex $COLORS [list $color_idx $segment_color]]
				}

				if {!$segment_color} {
					set segment_color {#555555}
				} {
					incr segment_color -1
					set segment_color [lindex $COLORS [list $color_idx $segment_color]]
				}
				$canvas_widget itemconfigure $leds($j,$i)	\
					-fill $segment_color -outline $outline
			}
		}
	}

	public method withdraw_window {} {
		wm withdraw $win
	}

	public method get_config {} {
		return [list		\
			$class_name	\
			[list		\
				[array get connection_port]	\
				[array get connection_pin]	\
				[wm geometry $win]		\
				[$canvas_widget.usr_note get]	\
				$conf_led_color			\
				$conf_dim_interval		\
			]	\
		]
	}

	public method set_config {state} {
		if {[catch {
			array set connection_port [lindex $state 0]
			array set connection_pin [lindex $state 1]
			if {[string length [lindex $state 2]]} {
				wm geometry $win [lindex $state 2]
			}
			
			$canvas_widget.usr_note delete 0
			$canvas_widget.usr_note insert 0 [lindex $state 3]
			
			set conf_led_color	[lindex $state 4]
			set conf_dim_interval	[lindex $state 5]
		
			for {set i 0} {$i < 8} {incr i} {
				set pin $connection_pin($i)
				if {$pin != {-}} {
					set pin	[expr {7 - $pin}]
				}
				set idx [lsearch -ascii -exact			\
					[$canvas_widget.cb_b$i cget -values]	\
					$pin					\
				]
				if {$idx == -1} {
					set idx 0
				}
				$canvas_widget.cb_b$i setvalue @$idx

				set idx [lsearch -ascii -exact			\
					[$canvas_widget.cb_p$i cget -values]	\
					$connection_port($i)			\
				]
				if {$idx == -1} {
					set idx 0
				}
				$canvas_widget.cb_p$i setvalue @$idx
			}
			
			for {set i 0} {$i < 4} {incr i} {
				set pin $connection_pin(T$i)
				if {$pin != {-}} {
					set pin	[expr {7 - $pin}]
				}
				set idx [lsearch -ascii -exact			\
					[$canvas_widget.cb_t_b$i cget -values]	\
					$pin					\
				]
				if {$idx == -1} {
					set idx 0
				}
				$canvas_widget.cb_t_b$i setvalue @$idx

				set idx [lsearch -ascii -exact			\
					[$canvas_widget.cb_t_p$i cget -values]	\
					$connection_port(T$i)			\
				]
				if {$idx == -1} {
					set idx 0
				}
				$canvas_widget.cb_t_p$i setvalue @$idx
			}

			new_state [$project pale_get_true_state]
			update

		}]} then {
			puts stderr "Unable to load config for $class_name"
			puts stderr $::errorInfo
			return 0
		} else {
			clear_modified
			return 1
		}
	}
	
	public method reset {} {
		dim_all
		new_state [$project pale_get_true_state]
	}

	# ------------------------------------------------------

	public method config_menu_special {} {
		set ::${class_name}::dim_interval $conf_dim_interval
		set ::${class_name}::color $conf_led_color
	}
	
	public method create_config_menu_special {} {
		foreach item {	Red	Orange	Yellow	Green	Blue	Purple	} \
			color {	#DD0000	#DD8800	#DDDD00	#00DD00	#0000DD	#8800DD	} \
		{
			$conf_menu.color entryconfigure $item -foreground $color
		}
	}

	public method dim_interval_changed {} {
		set conf_dim_interval ${::MultiplexedLedDisplay::dim_interval}
		set_modified
	}
	
	public method color_changed {} {
		set conf_led_color ${::MultiplexedLedDisplay::color}
	
		if {$drawing_on} {
			new_state [$project pale_get_true_state]
		}
		set_modified
	}
	
	public method dim_all {} {
		for {set i 0} {$i < 8} {incr i} {
			for {set j 0} {$j < 4} {incr j} {
				if {$prev_state($j,$i)} {
					set prev_state($j,$i) 0
					$canvas_widget itemconfigure $leds($j,$i)	\
						-fill {#555555} -outline {#FFFFFF}
				}
			}
		}
	}


	public method show_help_special {text_widget} {
# 		$text_widget insert insert "How to use this tool:\n"
# 		$text_widget tag add tag_bold insert-1l insert
# 		$text_widget insert insert "   1) Connect programmer to the computer\n"
# 		$text_widget tag add tag_bold insert-1l {insert-1l+6c}
# 		$text_widget insert insert "   2) Select port in section \"Programmer\"\n"
# 		$text_widget tag add tag_bold insert-1l {insert-1l+6c}
# 		$text_widget insert insert "   3) Click on button \"Autodetect device\"\n"
# 		$text_widget tag add tag_bold insert-1l {insert-1l+6c}
# 		$text_widget insert insert "   4) If programmer is operational you will see name of connected MCU in section \"Connected MCU\"\n"
# 		$text_widget tag add tag_bold insert-1l {insert-1l+6c}
# 		$text_widget insert insert "   5) Now you can perform arbitrary desired operation\n"
# 		$text_widget tag add tag_bold insert-1l {insert-1l+6c}
# 		$text_widget insert insert "\nMost useful functions are upload flash memory in section \"Flash memory (remember than flash must be first erased before programming). And \"Reprogram\" in section Special functions\", it will erase and reprogram all on-chip memory."
	}

	public method close_window_special {} {
	}
}
